import numpy as np
import skimage

from model_server.process import is_mask

def mask_largest_object(
        img: np.ndarray,
        max_allowed: int = 10,
        verbose: bool = True
) -> np.ndarray:
    """
    Where more than one connected component is found in an image, return the largest object by area
    :param img: (np.ndarray) containing object labels or binary mask
    :param max_allowed: raise an error if more than this number of objects is found
    :param verbose: print a message each time more than one object is found
    :return: np.ndarray of same size as img
    """
    if is_mask(img): # assign object labels
        ob_id = skimage.measure.label(img)
    else:  # assume img is contains object labels
        ob_id = img

    # import skimage
    # from pathlib import Path
    # where = Path('c:/Users/rhodes/projects/proj0011-plankton-seg/tmp')
    # skimage.io.imsave(where / 'raw.png', img)
    num_obj = len(np.unique(ob_id)) - 1
    if num_obj > max_allowed:
        raise TooManyObjectError(f'Found {num_obj} objects in frame')
    if num_obj > 1:
        if verbose:
            print(f'Found {num_obj} nonzero unique values in object map; keeping the one with the largest area')
        # pr = regionprops_table(ob_id, properties=['label', 'area'])
        val, cts = np.unique(ob_id, return_counts=True)
        mask = ob_id == val[1 + cts[1:].argmax()]
        # idx_max_area = pr['area'].argmax()
        # mask = ob_id == pr['label'][idx_max_area]
        return mask * img
    else:
        return img


class Error(Exception):
    pass


class TooManyObjectError(Exception):
    pass
